
`timescale 1ns / 100ps

module tb_top;


wire uart_rx ;
wire uart_tx ;


`define PARITY_OFF      1'b0
`define PARITY_ON       1'b1
`define PARITY_ODD      1'b0
`define PARITY_EVEN     1'b1
`define NSTOPS_1        1'b0
`define NSTOPS_2        1'b1
`define BAUD_115200     3'b000
`define BAUD_38400      3'b001
`define BAUD_28800      3'b010
`define BAUD_19200      3'b011
`define BAUD_9600       3'b100
`define BAUD_4800       3'b101
`define BAUD_2400       3'b110
`define BAUD_1200       3'b111
`define NBITS_7         1'b0
`define NBITS_8         1'b1

`define RECEIVE_RESULT_OK           4'b1000 
`define RECEIVE_RESULT_FALSESTART   4'b0001 
`define RECEIVE_RESULT_BADPARITY    4'b0010 
`define RECEIVE_RESULT_BADSTOP      4'b0100 


reg serial_out;    
reg serial_in;      
reg [7:0] get_serial_data;     
reg [7:0] get_serial_status;    


reg w_clk50m = 0;

top u_top (
    .i_clk(w_clk50m), 

    .o_led1() ,
    .o_led2() ,
    .o_led3() ,
    .o_led4() ,

    .o_tx(uart_tx),
    .i_rx(uart_rx)




);





// clock generator - 50MHz clock 
always #10 w_clk50m = ~w_clk50m;



assign uart_rx = serial_out;

always @(posedge w_clk50m)
   serial_in <= uart_tx;


initial 
begin 
    
    serial_out = 1;
	#1000000;  

    // command string: "W 0010 12345678" + LF 
	// command string: "R 0010" + LF 	
	// command string: "R 008A" + LF 
	// command string: "R 008B" + LF
	
	
    send_serial ("W", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0); 
    #100;
    send_serial (" ", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("1", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial (" ", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("1", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("2", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("3", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("4", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("5", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("6", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("7", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("8", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("\n", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    

	#1000000;
    send_serial ("R", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0); 
    #100;
    send_serial (" ", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("1", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
	#100;
    send_serial ("\n", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);	

	
	#1000000;
    send_serial ("R", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0); 
    #100;
    send_serial (" ", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("8", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("A", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
	#100;
    send_serial ("\n", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);	
	#10000000;
	
    send_serial ("R", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0); 
    #100;
    send_serial (" ", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("0", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("8", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
    #100;
    send_serial ("B", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);
	#100;
    send_serial ("\n", `BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8, 0);	
	#1000000;	
	

end 


initial 
begin 
    
    serial_in = 1;
    get_serial_data = 0;       
    get_serial_status = 0;
end 


always 
begin 

    get_serial(`BAUD_115200, `PARITY_EVEN, `PARITY_OFF, `NSTOPS_1, `NBITS_8);
    

    if (get_serial_status & `RECEIVE_RESULT_OK)
    begin
       
        if (get_serial_data >= 8'h19) 
            $display("received byte 0x%h (\"%c\") at %t ns", get_serial_data, get_serial_data, $time);
        else 
            $display("received byte 0x%h (\"%c\") at %t ns", get_serial_data, 8'hb0, $time);
    end 
    
   
    if (get_serial_status & `RECEIVE_RESULT_FALSESTART)
        $display("Error (get_char): false start condition at %t", $realtime);

          
    if (get_serial_status & `RECEIVE_RESULT_BADPARITY)
        $display("Error (get_char): bad parity condition at %t", $realtime);

   
    if (get_serial_status & `RECEIVE_RESULT_BADSTOP)
        $display("Error (get_char): bad stop bits sequence at %t", $realtime);
end 

//------------------------------------------------------------------






task send_serial;

input [7:0] inputchar;
input baud;
input paritytype;
input parityenable;
input nstops;
input nbits;
input baud_error_factor;

reg       nbits;
reg       parityenable;
reg       paritytype;
reg [1:0] baud;
reg       nstops;
integer     baud_error_factor; 
reg [7:0]   char;
reg [7:0]   disp_char;
reg         parity_bit;
integer     bit_time;
integer     num_of_bits;

begin 
    char = inputchar;
    parity_bit = 1'b0;
    
    case (baud)
        `BAUD_115200: bit_time = 1000000000/(115200 + 1152*baud_error_factor);
        `BAUD_38400:  bit_time = 1000000000/(38400 + 384*baud_error_factor);
        `BAUD_28800:  bit_time = 1000000000/(28800 + 288*baud_error_factor);
        `BAUD_19200:  bit_time = 1000000000/(19200 + 192*baud_error_factor);
        `BAUD_9600:   bit_time = 1000000000/(9600 + 96*baud_error_factor);
        `BAUD_4800:   bit_time = 1000000000/(4800 + 48*baud_error_factor);
        `BAUD_2400:   bit_time = 1000000000/(2400 + 24*baud_error_factor);
        `BAUD_1200:   bit_time = 1000000000/(1200 + 12*baud_error_factor);
    endcase   

    
    disp_char = (char >= 8'h19) ? char : 8'hb0;
    $display ("Sending character 0x%h (\"%c\"), at %0d baud (err=%0d), %0d bits, %0s parity, %0d stops",
        (nbits == `NBITS_7) ? (char & 8'h7f) : char,
        (nbits == `NBITS_7) ? (disp_char & 8'h7f) : disp_char,
        1000000000/bit_time,
        baud_error_factor,
        (nbits == `NBITS_7) ? 7 : 8,
        (parityenable == `PARITY_OFF) ? "NONE" : (paritytype == `PARITY_EVEN) ? "EVEN" : "ODD",
        (nstops == `NSTOPS_1) ? 1 : 2
    );

  
    serial_out = 1'b0;  
    #(bit_time);

    
    num_of_bits =  8;
    repeat (num_of_bits) 
    begin
        serial_out = char[0];
        #(bit_time);
        char = {1'b0, char[7:1]};
    end

    
    serial_out = 1'b1;   
    #(bit_time);

    if (nstops) 
        #(bit_time);
end
endtask


task get_serial;
// input parameters 
input baud;
input paritytype;
input parityenable;
input nstops;
input nbits;
// internal registers 
reg       nbits;
reg       parityenable;
reg       paritytype;
reg [1:0] baud;
reg       nstops;
integer     bit_time;
reg         expected_parity;
integer     num_of_bits;

begin 
 
    get_serial_status = 0;
    get_serial_data = 0;


    case (baud)
        `BAUD_115200: bit_time = 1000000000/115200;
        `BAUD_38400:  bit_time = 1000000000/38400;
        `BAUD_28800:  bit_time = 1000000000/28800;
        `BAUD_19200:  bit_time = 1000000000/19200;
        `BAUD_9600:   bit_time = 1000000000/9600;
        `BAUD_4800:   bit_time = 1000000000/4800;
        `BAUD_2400:   bit_time = 1000000000/2400;
        `BAUD_1200:   bit_time = 1000000000/1200;
    endcase   

    
    get_serial_status = `RECEIVE_RESULT_OK;  

    
    @(negedge serial_in);  

    
    #(bit_time/2);  

    
    if (serial_in != 0) 
        get_serial_status = get_serial_status | `RECEIVE_RESULT_FALSESTART;
    else 
    begin

        num_of_bits = 8;
        repeat (num_of_bits) 
        begin 
            
            #(bit_time);  
            
            get_serial_data = {serial_in, get_serial_data[7:1]};
        end



       
        #(bit_time);


            
		if (serial_in != 1)
			get_serial_status = get_serial_status | `RECEIVE_RESULT_BADSTOP;
		else
			
			if (nstops) 
			begin
				#(bit_time);
				if (serial_in != 1)
					get_serial_status = get_serial_status | `RECEIVE_RESULT_BADSTOP;
			end

    end
end
endtask






endmodule
